/*
 *
 *  $Id: pvrusb2-video-ivtv.c 830 2006-01-01 22:14:28Z isely $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*

   This source file is specifically designed to interface with the
   saa7115 support that is available in ivtv, which is entirely
   different than that which is available in v4l.

*/

#include "pvrusb2-video-ivtv.h"

#ifdef PVR2_ENABLE_SAA7115

#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-debug.h"
#include "decoder.h"
#include "ivtv.h"
#include <linux/videodev.h>
#include <linux/video_decoder.h>

#define MODE_UNKNOWN 0
#define MODE_OLD 1
#define MODE_NEW 2

struct pvr2_ivtv_decoder {
	struct pvr2_i2c_handler handler;
	struct pvr2_decoder_ctrl ctrl;
	struct pvr2_i2c_client *client;
	struct pvr2_hdw *hdw;
	int mode;
	unsigned long stale_mask;
};


static void set_input(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	int v = 0;
	pvr2_trace(PVR2_TRACE_CHIPS,"i2c ivtv set_input(%d)",
		   hdw->controls[PVR2_CID_INPUT].value);
	switch(hdw->controls[PVR2_CID_INPUT].value){
	case PVR2_CVAL_INPUT_TV:
		v = 4;
		break;
	case PVR2_CVAL_INPUT_COMPOSITE:
		v = 5;
		break;
	case PVR2_CVAL_INPUT_SVIDEO:
		v = 8;
		break;
	case PVR2_CVAL_INPUT_RADIO:
		// ????? No idea yet what to do here
	default:
		return;
	}
	pvr2_i2c_client_cmd(ctxt->client,DECODER_SET_INPUT,&v);
}


static int check_input(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	return hdw->controls[PVR2_CID_INPUT].dirty != 0;
}


static void set_bcsh(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	struct decoder_state st;
	int ret;
	pvr2_trace(PVR2_TRACE_CHIPS,
		   "i2c ivtv set_bcsh b=%d c=%d s=%d h=%d",
		   hdw->controls[PVR2_CID_BRIGHTNESS].value,
		   hdw->controls[PVR2_CID_CONTRAST].value,
		   hdw->controls[PVR2_CID_SATURATION].value,
		   hdw->controls[PVR2_CID_HUE].value);
	st.bright = hdw->controls[PVR2_CID_BRIGHTNESS].value;
	st.contrast = hdw->controls[PVR2_CID_CONTRAST].value;
	st.sat = hdw->controls[PVR2_CID_SATURATION].value;
	st.hue = hdw->controls[PVR2_CID_HUE].value;
	ret = -EINVAL;
	/* isely@pobox.com 15-Oct-2005 - Really ugly kludge to attempt to
	   operate with both older and newer types of saa7115 modules. */
	if ((ctxt->mode == MODE_UNKNOWN) || (ctxt->mode == MODE_NEW)) {
		ret = pvr2_i2c_client_cmd(ctxt->client,DECODER_SET_STATE,&st);
		if (!ret) {
			if (ctxt->mode == MODE_UNKNOWN) {
				pvr2_trace(PVR2_TRACE_INFO,
					   "newer saa7115 sensed");
				ctxt->mode = MODE_NEW;
			}
		}
		if (!ret || (ret != -EINVAL)) return;
	}
	if ((ctxt->mode == MODE_UNKNOWN) || (ctxt->mode == MODE_OLD)) {
		ret = pvr2_i2c_client_cmd(ctxt->client,
					  DECODER_SET_PICTURE,&st);
		if (!ret) {
			if (ctxt->mode == MODE_UNKNOWN) {
				pvr2_trace(PVR2_TRACE_INFO,
					   "old saa7115 sensed");
				ctxt->mode = MODE_OLD;
			}
		}
		if (!ret || (ret != -EINVAL)) return;
	}
}


static int check_bcsh(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	return (hdw->controls[PVR2_CID_BRIGHTNESS].dirty ||
		hdw->controls[PVR2_CID_CONTRAST].dirty ||
		hdw->controls[PVR2_CID_SATURATION].dirty ||
		hdw->controls[PVR2_CID_HUE].dirty);
}


static void set_std(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	int val;
	pvr2_trace(PVR2_TRACE_CHIPS,"i2c ivtv set_std %d",
		   hdw->controls[PVR2_CID_VIDEOSTANDARD].value);
	switch (hdw->controls[PVR2_CID_VIDEOSTANDARD].value) {
	default:
	case PVR2_CVAL_VIDEOSTANDARD_NTSC_M:
		val = VIDEO_MODE_NTSC;
		break;
	case PVR2_CVAL_VIDEOSTANDARD_SECAM_L:
		val = VIDEO_MODE_SECAM;
		break;
	case PVR2_CVAL_VIDEOSTANDARD_PAL_I:
	case PVR2_CVAL_VIDEOSTANDARD_PAL_DK:
	case PVR2_CVAL_VIDEOSTANDARD_PAL_BG:
		val = VIDEO_MODE_PAL;
		break;
	}
	pvr2_i2c_client_cmd(ctxt->client,DECODER_SET_NORM,&val);
}


static int check_std(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	return hdw->controls[PVR2_CID_VIDEOSTANDARD].dirty != 0;
}


static void set_audio(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	int val;
	pvr2_trace(PVR2_TRACE_CHIPS,"i2c ivtv set_audio %d",
		   hdw->controls[PVR2_CID_SRATE].value);
	switch (hdw->controls[PVR2_CID_SRATE].value) {
	default:
	case PVR2_CVAL_SRATE_48:
		val = DECODER_AUDIO_48_KHZ;
		break;
	case PVR2_CVAL_SRATE_44_1:
		val = DECODER_AUDIO_441_KHZ;
		break;
	}
	pvr2_i2c_client_cmd(ctxt->client,DECODER_SET_AUDIO,&val);
}


static int check_audio(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	return hdw->controls[PVR2_CID_SRATE].dirty != 0;
}


static void set_size(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	struct video_window wind;
	wind.width = hdw->controls[PVR2_CID_HRES].value;
	wind.height = hdw->controls[PVR2_CID_VRES].value;
	pvr2_trace(PVR2_TRACE_CHIPS,"pvr2_decoder_set_size(%dx%d)",
		   wind.width,wind.height);
	pvr2_i2c_client_cmd(ctxt->client,DECODER_SET_SIZE,&wind);
}


static int check_size(struct pvr2_ivtv_decoder *ctxt)
{
	struct pvr2_hdw *hdw = ctxt->hdw;
	return (hdw->controls[PVR2_CID_HRES].dirty ||
		hdw->controls[PVR2_CID_VRES].dirty);
}


struct pvr2_ivtv_decoder_ops {
	void (*update)(struct pvr2_ivtv_decoder *);
	int (*check)(struct pvr2_ivtv_decoder *);
};


static const struct pvr2_ivtv_decoder_ops decoder_ops[] = {
	{ .update = set_bcsh, .check = check_bcsh},
	{ .update = set_input, .check = check_input},
	{ .update = set_std, .check = check_std},
	{ .update = set_size, .check = check_size},
	{ .update = set_audio, .check = check_audio},
};


static void decoder_detach(struct pvr2_ivtv_decoder *ctxt)
{
	ctxt->client->handler = 0;
	ctxt->hdw->decoder_ctrl = 0;
	kfree(ctxt);
}


static int decoder_check(struct pvr2_ivtv_decoder *ctxt)
{
	unsigned long msk;
	unsigned int idx;

	for (idx = 0; idx < sizeof(decoder_ops)/sizeof(decoder_ops[0]); 
	     idx++) {
		msk = 1 << idx;
		if (ctxt->stale_mask & msk) continue;
		if (decoder_ops[idx].check(ctxt)) {
			ctxt->stale_mask |= msk;
		}
	}
	return ctxt->stale_mask != 0;
}


static void decoder_update(struct pvr2_ivtv_decoder *ctxt)
{
	unsigned long msk;
	unsigned int idx;

	for (idx = 0; idx < sizeof(decoder_ops)/sizeof(decoder_ops[0]); 
	     idx++) {
		msk = 1 << idx;
		if (!(ctxt->stale_mask & msk)) continue;
		ctxt->stale_mask &= ~msk;
		decoder_ops[idx].update(ctxt);
	}
}


static int decoder_detect(struct pvr2_i2c_client *cp)
{
	/* Attempt to query the decoder - let's see if it will answer */
	int status;
	int ret;

	ret = pvr2_i2c_client_cmd(cp,DECODER_GET_STATUS,&status);
	return ret == 0; /* Return true if it answered */
}


static void decoder_enable(struct pvr2_ivtv_decoder *ctxt,int fl)
{
	int val = fl ? 1 : 0;
	pvr2_trace(PVR2_TRACE_CHIPS,"i2c ivtv decoder_enable(%d)",val);
	pvr2_i2c_client_cmd(ctxt->client,DECODER_ENABLE_OUTPUT,&val);
}


static int decoder_is_tuned(struct pvr2_ivtv_decoder *ctxt)
{
	int status;
	int ret;
	ret = pvr2_i2c_client_cmd(ctxt->client,DECODER_GET_STATUS,&status);
	if (ret < 0) return -EINVAL;
	return (status & DECODER_STATUS_GOOD) ? 1 : 0;
}


const static struct pvr2_i2c_handler_functions hfuncs = {
	.detach = (void (*)(void *))decoder_detach,
	.check = (int (*)(void *))decoder_check,
	.update = (void (*)(void *))decoder_update,
};


int pvr2_i2c_decoder_ivtv_setup(struct pvr2_hdw *hdw,
				struct pvr2_i2c_client *cp)
{
	struct pvr2_ivtv_decoder *ctxt;

	if (hdw->decoder_ctrl) return 0;
	if (cp->handler) return 0;
	if (!decoder_detect(cp)) return 0;

	ctxt = kmalloc(sizeof(*ctxt),GFP_KERNEL);
	if (!ctxt) return 0;
	memset(ctxt,0,sizeof(*ctxt));

	ctxt->handler.func_data = ctxt;
	ctxt->handler.func_table = &hfuncs;
	ctxt->ctrl.ctxt = ctxt;
	ctxt->ctrl.detach = (void (*)(void *))decoder_detach;
	ctxt->ctrl.enable = (void (*)(void *,int))decoder_enable;
	ctxt->ctrl.tuned = (int (*)(void *))decoder_is_tuned;
	ctxt->client = cp;
	ctxt->hdw = hdw;
	ctxt->mode = MODE_UNKNOWN;
	ctxt->stale_mask = (1 << (sizeof(decoder_ops)/
				  sizeof(decoder_ops[0]))) - 1;
	hdw->decoder_ctrl = &ctxt->ctrl;
	cp->handler = &ctxt->handler;
	return !0;
}


#endif /* PVR2_ENABLE_SAA7115 */

/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 70 ***
  *** tab-width: 8 ***
  *** c-basic-offset: 8 ***
  *** End: ***
  */
